// dllmain.cpp : Defines the entry point for the DLL application.
#include "pch.h"
#include "dllmain.h"
#include <string>
#include <vector>
#include "osc/OscReceivedElements.h"
#include "osc/OscPacketListener.h"
#include "ip/UdpSocket.h"

class ExamplePacketListener : public osc::OscPacketListener {
protected:

    virtual void ProcessMessage(const osc::ReceivedMessage& m,
        const IpEndpointName& remoteEndpoint)
    {
        OutputDebugString(L"Got OSC Message!\r\n");
        (void)remoteEndpoint; // suppress unused parameter warning

        try {
            // example of parsing single messages. osc::OsckPacketListener
            // handles the bundle traversal.

            if (std::strcmp(m.AddressPattern(), "/test1") == 0) {
                // example #1 -- argument stream interface
                osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
                bool a1;
                osc::int32 a2;
                float a3;
                const char* a4;
                args >> a1 >> a2 >> a3 >> a4 >> osc::EndMessage;
                OutputDebugString(L"received '/test1' message with arguments: ... ");
                //std::cout << "received '/test1' message with arguments: "
                //    << a1 << " " << a2 << " " << a3 << " " << a4 << "\n";

            }
            else if (std::strcmp(m.AddressPattern(), "/test2") == 0) {
                // example #2 -- argument iterator interface, supports
                // reflection for overloaded messages (eg you can call 
                // (*arg)->IsBool() to check if a bool was passed etc).
                osc::ReceivedMessage::const_iterator arg = m.ArgumentsBegin();
                //bool a1 = (arg++)->AsBool();
                //int a2 = (arg++)->AsInt32();
                //float a3 = (arg++)->AsFloat();
                const char* a4 = (arg++)->AsString();
                //if (arg != m.ArgumentsEnd())
                    //throw osc::ExcessArgumentException();

                OutputDebugString(L"received '/test2' message with argument: ");
                OutputDebugStringA(a4);
                OutputDebugString(L"\r\n");
                //std::cout << "received '/test2' message with arguments: "
                //    << a1 << " " << a2 << " " << a3 << " " << a4 << "\n";
            }
        }
        catch (osc::Exception& e) {
            // any parsing errors such as unexpected argument types, or 
            // missing arguments get thrown as exceptions.
            OutputDebugString(L"error while parsing message: ...");
            //std::cout << "error while parsing message: "
            //    << m.AddressPattern() << ": " << e.what() << "\n";
        }
    }
};

/* Definition and declaration of the OSC Listeners. 
* Contains OSC addresses to listen to and OS message ID to send for each 
* address */
struct Listener {
    std::string address;
    unsigned int dataType;
    unsigned int message;
};
std::vector<Listener> listeners;
/* Handle to the calling window */
HWND hwnd = NULL;
ExamplePacketListener* examplePacketListener;
UdpListeningReceiveSocket* sock;

BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    switch (ul_reason_for_call)
    {
    case DLL_PROCESS_ATTACH:
    case DLL_THREAD_ATTACH:
        OutputDebugString(L"ATTACH\r\n");
        break;
    case DLL_THREAD_DETACH:
    case DLL_PROCESS_DETACH:
        close();
        break;
    }
    return TRUE;
}

DLLEXPORT int open(HWND targetWindowHandle, unsigned int port)
{
    hwnd = targetWindowHandle;
    examplePacketListener = new ExamplePacketListener;
    sock =  new UdpListeningReceiveSocket(
        IpEndpointName(IpEndpointName::ANY_ADDRESS, port),
        examplePacketListener);
    sock->RunUntilSigInt();
    OutputDebugString(L"open\r\n");
    return 0;
}

DLLEXPORT int close()
{
    hwnd = NULL;
    listeners.clear();
    return 0;
}

DLLEXPORT int addListener(std::string address_, unsigned int messageID_, unsigned int dataType_)
{
    listeners.push_back(Listener{ address_, dataType_, messageID_ });
    return 0;
}

DLLEXPORT int removeListener(std::string address_)
{
    UINT result = 1;

    //All OSC addresses have to start with a '/'
    if (address_[0] != '/') address_.insert(0, "/");

    for (UINT i = 0; i < listeners.size(); i++)
    {
        if (listeners[i].address == address_)
        {
            listeners.erase(listeners.begin() + i);
            i--; //one step backward, what previously was i+1 is i after erase().
            result = 0; //Found at least one occurence
        }
    }
    return result;
}

/* Debugging function to test messaging. Probably will be removed later. */
DLLEXPORT int testMsg(HWND windowHandle, unsigned int messageID)
{
    if (PostMessage(windowHandle, messageID, 42, 43))
        return 0;
    else
        return 1;
}
